// dbclass - ODBC/SQL database wrapper class, (c) 2000 by Tom Grandgent
//
// This code may not be used in commercial products without the express 
// written consent of Tom Grandgent (tom@grandgent.com)
//
// 11/15/2000 - last major modifications
//
// 7/31/2001 - moved sample code into a separate project

#include <windows.h>
#include <sql.h>
#include <sqlext.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "dbclass.h"

// Plain constructor, no opening of a datasource or database file
Database::Database(void)
{
	init_odbc();
	if (error)
		return;
}

// Plain constructor
Recordset::Recordset(Database &db)
{
	// Save parent database
	parent = &db;
	// Initialize various things to NULL
	hstmt = NULL;
	sql_statement = NULL;

}

// Destructor - cleans up any remaining query stuff
Recordset::~Recordset(void)
{

	// Close SQL statement (if one is open)
	if (hstmt)
	{
		// This next line of code crashes ODBCJT32.DLL for some reason.  I don't think it likes being in a destructor.
		// No worries though - any statement handles will be freed automatically when the ODBC environment is closed.
		//SQLFreeStmt(hstmt, SQL_DROP);
		hstmt = NULL;
	}
	// Frees the SQL statement text (if there is any)
	if (sql_statement)
	{
		free(sql_statement);
		sql_statement = NULL;
	}
}

// Constructor with a datasource or database name
Database::Database(char *sourcename)
{
	init_odbc();
	if (error)
		return;

	open(sourcename);
}

// Constructor with a datasource or database name and username/password
Database::Database(char *sourcename, char *username, char *password)
{
	init_odbc();
	if (error)
		return;

	open(sourcename, username, password);
}

// Destructor - cleans up datasource and ODBC environment
Database::~Database(void)
{
	close();
	shutdown_odbc();
}

// Set the app-defined error handler function
void Database::set_error_handler(void (*a_error_handler)(char *text))
{
	error_handler = a_error_handler;
}

// Default error handler - set in init_odbc()
void default_error_handler(char *text)
{
	//MessageBox(NULL, text, "SQL/ODBC Error", MB_ICONERROR);
	printf("%s\n", text);
}

// Initializes the ODBC environment and connection to the datasource
void Database::init_odbc(void)
{
	set_error_handler(default_error_handler);
	datasource[0] = '\0';
	username[0] = '\0';
	password[0] = '\0';
	cbErrorMsgMax = SQL_MAX_MESSAGE_LENGTH - 1;
	henv = hdbc = NULL;

	error = SQLAllocEnv(&henv); /* Environment handle */ 
}

// Shuts down the ODBC environment
void Database::shutdown_odbc(void)
{
	// Close ODBC environment
	if (henv)
	{
		SQLFreeHandle(SQL_HANDLE_ENV, henv);
		henv = NULL;
	}
}

// Handle an error, getting details and calling the error handler (if one exists)
void Database::handle_error(SQLHSTMT hstmt, RETCODE error, char *text, ...)
{
	va_list ap;
	static char s[4096];

	if (!error_handler)
		return;
	
	va_start(ap, text);
	vsprintf(s, text, ap);
	va_end(ap);

	if (!SQL_SUCCEEDED(error))
	{
		SQLError(henv, hdbc, hstmt, 
			szSqlState, &pfNativeError, szErrorMsg,
			cbErrorMsgMax, &cbErrorMsgMax);
		strcat(s, "\n");
		sprintf(&s[strlen(s)], "SQLError() reports: SqlState '%s', NativeError '%d', ErrorMsg '%s'\n", szSqlState, pfNativeError, szErrorMsg);
	}

	error_handler(s);
}

// Handles an error for a Recordset by using its parent database error handler
void Recordset::handle_error(char *text, ...)
{
	va_list ap;
	static char s[4096];

	va_start(ap, text);
	vsprintf(s, text, ap);
	va_end(ap);

	// Call the error handler for the parent Database object
	parent->handle_error(hstmt, error, s);
}

// Opens the given datasource or database with the given username and password
int Database::open(char *a_sourcename, char *a_username, char *a_password)
{
	strcpy(username, a_username);
	strcpy(password, a_password);
	return open(a_sourcename);
}

// Opens the given datasource or database
int Database::open(char *a_sourcename)
{
	char *ch;
	char connect_string[256];
	unsigned char szConnStr[256];
	short int cbConnStr;

	strcpy(datasource, a_sourcename);

	__try
	{
		error = SQLAllocConnect(henv, &hdbc); /* Connection handle */ 
		if (!SQL_SUCCEEDED(error))
			return error;

		/* Set login timeout to 5 seconds. */ 
		SQLSetConnectOption(hdbc, SQL_LOGIN_TIMEOUT, 5); 

		/* Connect to data source */ 
		//printf("Trying to connect to '%s' with l/p '%s'/'%s'\n", datasource, username, password);
		// See if the given datasource is actually a filename (check for an extension)
		ch = strrchr(datasource, '.');
		if (ch)
		{
			// Try to identify the driver based on the extension
			if (!stricmp(ch, ".mdb"))
				sprintf(connect_string, "DBQ=%s;DRIVER={Microsoft Access Driver (*.mdb)}", datasource);
			else if (!stricmp(ch, ".xls"))
				sprintf(connect_string, "DBQ=%s;DRIVER={Microsoft Excel Driver (*.xls)}", datasource);
			else if (!stricmp(ch, ".txt") || !stricmp(ch, ".csv"))
				sprintf(connect_string, "DBQ=%s;DRIVER={Microsoft Text Driver (*.txt; *.csv)}", datasource);
			else
				ch = NULL;
			// Open the database file if its extension was recognized
			if (ch)
			{
				error = SQLDriverConnect(hdbc, NULL, (unsigned char *) connect_string, 
					strlen(connect_string), szConnStr, sizeof(szConnStr), 
					&cbConnStr, SQL_DRIVER_NOPROMPT);
				if (!SQL_SUCCEEDED(error))
					return error;
			}
		}
		if (!ch)
			error = SQLConnect(hdbc, (unsigned char *) datasource, SQL_NTS, (unsigned char *) username, SQL_NTS, (unsigned char *) password, SQL_NTS);
	}
	__finally
	{
		if (!SQL_SUCCEEDED(error))
			handle_error(NULL, error, "Error opening '%s'\n", datasource);
	}

	return error;
}

// Closes the database
void Database::close(void)
{
	// Close database connection
	if (hdbc)
	{
		SQLDisconnect(hdbc);
		SQLFreeHandle(SQL_HANDLE_DBC, hdbc);
		hdbc = NULL;
	}
}

// Sets up a query - stores the SQL statement and other info
int Recordset::query(char *a_sql_statement, ...)
{
	va_list ap;
	static char s[4096];

	// Parse the variable arguments and come up with one string
	va_start(ap, a_sql_statement);
	vsprintf(s, a_sql_statement, ap);
	va_end(ap);

	// If there's a previous SQL statement, free it
	if (sql_statement)
	{
		free(sql_statement);
		sql_statement = NULL;
	}
	
	// Save parameters to this function
	sql_statement = strdup(s);
	query_type = QUERY_WRITE;

//#ifdef _DEBUG
//	Log("query(%s)", sql_statement);
//#endif

	// Figure out what kind of query this is from the beginning of the SQL
	if (!strnicmp(sql_statement, "SELECT", 6))
		query_type = QUERY_READ;

	// Reset state parameters from a previous query
	parameters = 0;
	num_columns = 0;
	cols_enumerated = 0;

	// If there's a previous SQL statement handle, close it
	if (hstmt)
		SQLCloseCursor(hstmt);
	else
		error = SQLAllocStmt(parent->hdbc, &hstmt); // Allocate a new statement handle

	switch (query_type)
	{
		case QUERY_WRITE:
			error = SQLPrepare(hstmt, (unsigned char *) sql_statement, SQL_NTS);
			break;
		case QUERY_READ:
			error = SQLExecDirect(hstmt, (unsigned char *) sql_statement, SQL_NTS);
			break;
		default:
			error = 1;
			handle_error("Unknown query type when beginning a query: %d", query_type);
			return error;
	}

	// Get the number of columns for this query
	if (SQL_SUCCEEDED(error))
		SQLNumResultCols(hstmt, &num_columns);
	else
		handle_error("Error in query()");

	return error;
}

// Sets up a parameter for a query
int Recordset::parameter(short int datatype, void *dataptr, int datalen)
{
	// This is a new parameter, increment the # of parameters (starts at 1)
	parameters++;

	return parameter_column(parameters, datatype, dataptr, datalen);
}

// Sets up a parameter for a query given a column name
int Recordset::named_parameter(char *name, short int datatype, void *dataptr, int datalen)
{
	int i;	// This can't be static for some stupid reason???

	// Enumerate the columns if they haven't been already
	if (!cols_enumerated)
		enumerate_columns();

	// Look through the enumerated column names for a match
	for (i = 0; i < num_columns; i++)
		if (!stricmp(col_name[i], name))
			break;

	// Bail if no match was found
	if (i == num_columns)
	{
		handle_error("named_parameter() failure: Column '%s' not found.", name);
		return SQL_ERROR;
	}
	i++;
	return parameter_column(i, datatype, dataptr, datalen);
}

// Enumerates the columns (names, data types, etc) for this query
int Recordset::enumerate_columns(void)
{
	int i;
	SQLSMALLINT name_len;

	for (i = 0; i < ((num_columns < DBCLASS_MAX_COLS) ? num_columns : DBCLASS_MAX_COLS - 1); i++)
	{
		error = SQLDescribeCol(hstmt, i + 1, (unsigned char *) col_name[i], 
			DBCLASS_MAX_COL_NAME_LENGTH, &name_len, &col_type[i], 
			&col_size[i], &col_decimal[i], &col_nullable[i]);
		if (!SQL_SUCCEEDED(error))
		{
			handle_error("Error enumerating column #%d", i);
			return SQL_ERROR;
		}
		//printf("Column %d: %s\n", i, col_name[i]);
	}

	cols_enumerated = 1;

	return 0;
}

// Sets up a parameter for a query with an explicitly specified column number (starts at 1)
int Recordset::parameter_column(int column, short int datatype, void *dataptr, int datalen)
{
	long i;

	switch (query_type)
	{
		case QUERY_WRITE:
			error = SQLBindParameter(hstmt, column, SQL_PARAM_INPUT, datatype, datatype, datalen, 0, dataptr, 0, NULL);
			if (!SQL_SUCCEEDED(error))
				handle_error("Error binding write-query column #%d", column);
			break;
		case QUERY_READ:
			i = (long) datalen;
			//if (i)
			//	SQLBindCol(hstmt, column, datatype, dataptr, datalen, &i);
			//else
				SQLBindCol(hstmt, column, datatype, dataptr, datalen, NULL);
			if (!SQL_SUCCEEDED(error))
				handle_error("Error binding read-query column #%d", column);
			break;
	}

	return error;
}

// Executes a query - can be called multiple times for a single query
// It is the caller's responsibility to check for an error using SQL_SUCCEEDED()!
int Recordset::execute_query(void)
{
	switch (query_type)
	{
		case QUERY_WRITE:
			error = SQLExecute(hstmt);
//			if (!SQL_SUCCEEDED(error))
//				handle_error("Error executing write query (SQLExecute())");
			break;
		case QUERY_READ:
			error = SQLFetch(hstmt);
//			if (!SQL_SUCCEEDED(error))
//				handle_error("Error executing read query (SQLFetch())");
			break;
	}
	
	return error;
}

// Returns number of rows affected by a query, or -1 if N/A
int Recordset::row_count(void)
{
	SQLINTEGER i;

	error = SQLRowCount(hstmt, &i);
	if (!SQL_SUCCEEDED(error))
	{
		handle_error("Error getting row count (SQLRowCount())");
		return 0;
	}

	return (int) i;
}

// Returns number of columns in a query
int Recordset::col_count(void)
{
	return num_columns;
}
